(* $Id: pax.ml 752 2008-05-23 00:41:05Z gerd $ *)

open Pax_types
open Printf

type in_archive =
    { in_ar : archive;
      in_next : unit -> archive_rd option;
      in_close : unit -> unit;
      in_postprocess : (string, string->unit) Hashtbl.t
    }

type in_element =
 in_archive * archive_rd


type in_file =
    [ `File of string
    | `Stdin
    | `Obj_channel of Netchannels.in_obj_channel
    ]


type compression_kind =
    [ `None
    | `Gzip
    | `Bzip2
    ]

type format =
    [ `Tar
    ]

type overwrite_mode =
    [ `Always
    | `Update
    | `Ignore
    | `Reject
    ]

let gzip_cmd() =
  if !Godi_boot.is_bootconsole then
    "gzip"
  else
    Godi_conf.cmd_gzip

let bzip2_cmd() =
  Godi_conf.cmd_bzip2


let mk_gzip_cmd filename_opt  =
  sprintf "%s -c >%s"
  (gzip_cmd())
    (match filename_opt with
       | None -> ""
       | Some name -> 
	   let name' = Filename.quote name in
	   if name <> "" && name.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )



let mk_gunzip_cmd filename_opt =
  sprintf "%s -cd %s"
    (gzip_cmd())
    (match filename_opt with
       | None -> ""
       | Some name -> 
	   let name' = Filename.quote name in
	   if name <> "" && name.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )


let mk_bzip2_cmd filename_opt  =
  sprintf "%s -c >%s"
    (bzip2_cmd())
    (match filename_opt with
       | None -> ""
       | Some name -> 
	   let name' = Filename.quote name in
	   if name <> "" && name.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )



let mk_bunzip2_cmd filename_opt =
  sprintf "%s -cd %s"
    (bzip2_cmd())
    (match filename_opt with
       | None -> ""
       | Some name -> 
	   let name' = Filename.quote name in
	   if name <> "" && name.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )


let copy_channel 
      len 
      (src_ch : Netchannels.in_obj_channel) 
      (dest_ch : Netchannels.out_obj_channel) =
  (* Copies contents from src_ch to dest_ch. Returns [true] if at EOF.
   *)
  let buf = String.create 4096 in
  let slen = String.length buf in
  let k = ref 0L in
  try
    while true do
      let m = Int64.to_int (min (Int64.of_int slen) (Int64.sub len !k)) in
      if m = 0 then raise Exit;
      let n = src_ch # input buf 0 m in
      if n = 0 then raise Sys_blocked_io;
      dest_ch # really_output buf 0 n;
      k := Int64.add !k (Int64.of_int n)
    done;
    assert false
  with
    | End_of_file
    | Exit ->
	()




let read in_file compression fmt =
  let (ch, close) =
    match (in_file, compression) with
      | (`File name), `None ->
	  let ch = new Netchannels.input_channel (open_in_bin name) in
	  (ch, ch#close_in)
      | `Stdin, `None ->
	  let ch = 
	    Netchannels.lift_in 
	      (`Raw(new Netchannels.input_descr Unix.stdin)) in
	  (ch, (fun () -> Pax_tar.skip ch))
      | (`Obj_channel ch), `None ->
	  (ch, ch#close_in)
      | (`File name), `Gzip ->
	  let ch = 
	    new Netchannels.input_command (mk_gunzip_cmd (Some name)) in
	  (ch, (fun () -> Pax_tar.skip ch; ch#close_in()))
      | `Stdin,`Gzip ->
	  let ch = new Netchannels.input_command (mk_gunzip_cmd None) in
	  (ch, (fun () -> Pax_tar.skip ch; ch#close_in()))
      | (`File name), `Bzip2 ->
	  let ch = 
	    new Netchannels.input_command (mk_bunzip2_cmd (Some name)) in
	  (ch, (fun () -> Pax_tar.skip ch; ch#close_in()))
      | `Stdin,`Bzip2 ->
	  let ch = new Netchannels.input_command (mk_bunzip2_cmd None) in
	  (ch, (fun () -> Pax_tar.skip ch; ch#close_in()))
      | (`Obj_channel _), (`Gzip | `Bzip2) ->
	  failwith 
	    "Pax.read: object channels are incompatible with compression"
  in
  let ar_file =
    ( object
	method read_file() =
	  ch
	method create_file() =
	  assert false
      end
    ) in
  let ar =
    match fmt with
      | `Tar ->
	  new Pax_tar.tar ar_file in
  let cur_el_opt = ref (lazy(ar # read_archive())) in
  { in_ar = ar;
    in_next = (fun () ->
		 match Lazy.force !cur_el_opt with
		   | None -> None
		   | Some el -> 
		       cur_el_opt := lazy (el # next());
		       Some el
	      );
    in_postprocess = Hashtbl.create 50;
    in_close = close
  }


let read_element in_ar =
  match in_ar.in_next() with
    | None -> None
    | Some in_el -> Some (in_ar, in_el)


let get_metadata (_,in_el) =
  in_el # get_metadata

let extract_element_to_memory (_, in_el) =
  let buf = Buffer.create 1000 in
  let buf_ch = new Netchannels.output_buffer buf in
  let data_ch,_ = in_el # get_data() in
  buf_ch # output_channel data_ch;
  Buffer.contents buf

let slash_re = Str.regexp "/+"

let is_bad_pathname name =
  let l = Str.split_delim slash_re name in
  (l <> [] && List.hd l = "") ||
    (List.mem ".." l)


let iter_dirs f1 f2 split_name =
  let rec iter acc split_name =
    match split_name with
      | [] ->
	  ()
      | [ last_name ]
      | [ last_name; "" ] ->
	  f2 (String.concat "/" (List.rev (last_name :: acc)))
      | inner_name :: split_name' ->
	  f1 (String.concat "/" (List.rev (inner_name :: acc)));
	  iter (inner_name :: acc) split_name'
  in
  match split_name with
    | "" :: split_name' ->
	iter [ "" ] split_name'
    | _ ->
	iter [] split_name


let create_dir dirpath =
  if not(Godi_file.file_is_dir dirpath) then (
    if Godi_file.file_exists dirpath then
      failwith ("Not a directory: " ^ dirpath);
    Godi_file.create_dir dirpath
  )

let set_mtime path mtime =
  (* TODO: Move to Godi_file *)
  Unix.utimes path 0.0 (Int64.to_float mtime)


let norm_mode_bits is_dir bits =
  (* We look only at the user bits, and support only the combinations:
     r--: read-only, non-executable  (not for directories)
     rw-: read-write, non-executable (not for directories)
     r-x: read-only, executable      (not for directories)
     rwx: read-write, executable

     The bits for group and worls are set to this, too, and it is expected
     that the umask deletes bits.
   *)
  let is_writable = is_dir || (bits land 0o200) <> 0 in
  let is_executable = is_dir || (bits land 0o100) <> 0 in
  let b = 
    4 lor
      (if is_writable then 2 else 0) lor
      (if is_executable then 1 else 0) in
  (b lsl 6) lor (b lsl 3) lor b


let extract_element_as_file ?(insecure = false)
                            ?(prefer_link = false)
                            ?prefix 
                            ?(subst = fun s -> s) 
                            ?(preserve_mtime = false)
                            ?(overwrite = `Always)
                            (in_ar, in_el) =
  let m = in_el # get_metadata in
  let ok =
    match m # filetype with
      | `Regular -> true
      | `Directory -> true
      | `Hardlink _ -> true
      | `Symlink _ -> true
      | _ -> false in
  if ok then (
    if not insecure && is_bad_pathname m#name then
      failwith ("Refusing to extract this path name: " ^ m#name);
    let mapped_name = subst m#name in
    if mapped_name <> "" then (
      let local_name =
	match prefix with
	  | None -> mapped_name
	  | Some p -> 
	      if mapped_name.[0] = '/' then 
		mapped_name 
	      else
		Filename.concat p mapped_name in
      let split_name = Str.split_delim slash_re local_name in
      iter_dirs
	create_dir
	(fun filepath ->
	   match m # filetype with
	     | `Directory ->
		 create_dir filepath;
		 
		 if preserve_mtime then (
		   Hashtbl.replace 
		     in_ar.in_postprocess
		     filepath
		     (fun p -> set_mtime p m#mtime)
		 )
		     
	     | `Regular
	     | `Hardlink _
	     | `Symlink _ ->

		 let in_ch, file_id_opt = in_el#get_data() in

		 let is_self_copy =
		   ( match file_id_opt with
		       | Some (_, src_file_id) ->
			   if Sys.file_exists filepath then
			     let dest_file_id = Godi_file.file_id filepath in
			     src_file_id = dest_file_id
			   else
			     false 
		       | None ->
			   false
		   ) in

		 let do_overwrite, do_unlink =
		   if is_self_copy then (
		     prerr_endline ("Refusing to overwrite file by itself: " ^ 
				      filepath);
		     (false, false)
		   )
		   else
		     if Sys.file_exists filepath then
		       match overwrite with
			 | `Always -> 
			     (true,true)
			 | `Update ->
			     let t = Godi_file.file_mtime filepath in
			     if  t < Int64.to_float m#mtime then
			       (true,true)
			     else
			       (false,false)
			 | `Ignore -> 
			     (false,false) 
			 | `Reject -> 
			     failwith("File exists: " ^ filepath)
		     else
		       (true, false) in
		 if do_overwrite then (
		   if do_unlink then
		     Sys.remove filepath;
		   
		   ( match m # filetype with
		       | `Regular ->
			   let do_copy = ref true in
			   if prefer_link && file_id_opt <> None then (
			     match file_id_opt with
			       | Some(srcname,_) ->
				   ( try (* TODO: Use Godi_file *)
				       Unix.link srcname filepath;
				       do_copy := false
				     with _ -> ()
				   )
			       | None -> assert false
			   );
			   if !do_copy then (
			     let filemode = norm_mode_bits false m#mode in
			     let ch = 
			       open_out_gen
				 [Open_wronly; Open_creat; Open_binary]
				 filemode
				 filepath in
			     let out_ch = new Netchannels.output_channel ch in
			     copy_channel m#size in_ch out_ch;
			     out_ch # close_out();
			   
			     if preserve_mtime then
			       set_mtime filepath m#mtime
			   )
			       
		       | `Hardlink link ->
			   (* TODO: Use Godi_file *)
			   Unix.link link filepath
			     
		       | `Symlink link ->
			   Godi_file.create_symlink link filepath

		       | _ ->
			   ()
		   );
		 );
		 in_ch # close_in();
	     | _ -> 
		 assert false
	)
	split_name
    )

  )


let postprocess_ht ht =
  Hashtbl.iter
    (fun path action ->
       action path
    )
    ht;
  Hashtbl.clear ht


let postprocess_extraction in_ar =
  postprocess_ht in_ar.in_postprocess


let close_in in_ar =
  in_ar.in_close()


type out_archive = 
    { out_ar : archive;
      out_wr : archive_wr;
      out_close : unit -> unit
    }

type out_file =
    [ `File of string
    | `Stdout
    | `Obj_channel of Netchannels.out_obj_channel
    ]

let create out_file compression fmt =
  let (ch, close) =
    match (out_file, compression) with
      | (`File name), `None ->
	  let ch = new Netchannels.output_channel (open_out_bin name) in
	  (ch, ch#close_out)
      | `Stdout, `None ->
	  let ch = 
	    Netchannels.lift_out 
	      (`Raw(new Netchannels.output_descr Unix.stdout)) in
	  (ch, (fun () -> ()))
      | (`Obj_channel ch), `None ->
	  (ch, ch#close_out)
      | (`File name), `Gzip ->
	  let ch = 
	    new Netchannels.output_command (mk_gzip_cmd (Some name)) in
	  (ch, ch#close_out)
      | `Stdout,`Gzip ->
	  let ch = new Netchannels.output_command (mk_gzip_cmd None) in
	  (ch, ch#close_out)
      | (`File name), `Bzip2 ->
	  let ch = 
	    new Netchannels.output_command (mk_bzip2_cmd (Some name)) in
	  (ch, ch#close_out)
      | `Stdout,`Bzip2 ->
	  let ch = new Netchannels.output_command (mk_bzip2_cmd None) in
	  (ch, ch#close_out)
      | (`Obj_channel _), (`Gzip | `Bzip2) ->
	  failwith 
	    "Pax.create: object channels are incompatible with compression"
  in
  let ar_file =
    ( object
	method read_file() =
	  assert false
	method create_file() =
	  ch
      end
    ) in
  let ar =
    match fmt with
      | `Tar ->
	  new Pax_tar.tar ar_file in
  let ar_wr = ar # create_archive() in
  { out_ar = ar;
    out_wr = ar_wr;
    out_close = close
  }

let close_out out_ar =
  out_ar.out_wr # close();
  out_ar.out_close()

let add_element_from_file ?prefix ?(subst=fun s -> s) out_ar filename =
  let subst_name = subst filename in
  if subst_name <> "" then (
    let local_name =
      match prefix with
	| None -> filename
	| Some p -> 
	    if filename <> "" && filename.[0] = '/' then 
	      filename 
	    else
	      Filename.concat p filename in
    let st = Unix.LargeFile.lstat local_name in  (* TODO: move to Godi_file *)
    let filetype_opt =
      match st.Unix.LargeFile.st_kind with
	| Unix.S_REG ->
	    Some `Regular
	| Unix.S_DIR ->
	    Some `Directory
	| Unix.S_LNK ->
	    Some (`Symlink (Unix.readlink local_name))
	| _ ->
	    None in
    match filetype_opt with
      | None ->
	  ()   (* Ignore these *)
      | Some filetype ->
	  let m =
	    ( object 
		method name = subst_name
		method mode = st.Unix.LargeFile.st_perm
		method size = st.Unix.LargeFile.st_size
		method owner_uid = None
		method owner_uname = None
		method owner_gid = None
		method owner_gname = None
		method mtime = Int64.of_float st.Unix.LargeFile.st_mtime
		method filetype = filetype
		method format = None
		method have_data = false  (* ignored anyway *)
	      end : metadata
	    ) in
	  if filetype = `Regular && out_ar.out_ar#archive_type <> "clipboard" 
	  then (
	    let data_ch = 
	      out_ar.out_wr # put_data m in
	    let file_ch = 
	      new Netchannels.input_channel (open_in_bin local_name) in
	    copy_channel m#size file_ch data_ch;
	    data_ch # close_out();
	  )
	  else
	    out_ar.out_wr # put_metadata_only m
  )


let add_element_from_string ~name ~mode ~mtime ~filetype out_ar data =
  let m =
    ( object 
	method name = name
	method mode = mode
	method size = Int64.of_int (String.length data)
	method owner_uid = None
	method owner_uname = None
	method owner_gid = None
	method owner_gname = None
	method mtime = mtime
	method filetype = filetype
	method format = None
	method have_data = false  (* ignored anyway *)
      end : metadata
    ) in
  let data_ch = out_ar.out_wr # put_data m in
  data_ch # output_string data;
  data_ch # close_out()

  
type file_iterator = unit -> string option


let expand_directories ?prefix (f : file_iterator) =

  let current_fetcher = ref (fun () -> assert false) in

  let rec fetch_from_dir localname filename followup =
    let files = ref (Godi_file.read_dir localname) in
    fun () ->
      match !files with
	| file :: files' ->
	    files := files';
	    let fname = Filename.concat filename file in
	    let lname = Filename.concat localname file in
	    if Godi_file.file_is_dir lname then (
	      let old_fetcher = !current_fetcher in
	      current_fetcher := fetch_from_dir lname fname old_fetcher;
	      Some fname
	    )
	    else
	      Some fname
	| [] ->
	    current_fetcher := followup;
	    followup()
  in

  let fetch_from f followup () =
    let filename_opt = f() in
    match filename_opt with
      | None ->
	  current_fetcher := followup;
	  followup()
      | Some filename ->
	  let localname =
	    match prefix with
	      | None -> filename
	      | Some p -> 
		  if filename <> "" && filename.[0] = '/' then 
		    filename 
		  else
		    Filename.concat p filename in
	  if Godi_file.file_type ~is_dir:true ~is_not_symlink:true localname then (
	    let old_fetcher = !current_fetcher in
	    current_fetcher := fetch_from_dir localname filename old_fetcher;
	    Some filename
	  )
	  else
	    Some filename
  in

  let return_end() =
    None
  in

  current_fetcher := fetch_from f return_end;

  (fun () ->
     !current_fetcher()
  )


let file_iterator_of_list ?(nodescent=false) ?prefix list =
  let l = ref list in
  let f () =
    match !l with
      | name :: l' ->
	  l := l';
	  Some name
      | [] ->
	  None in
  if nodescent then
    f
  else
    expand_directories ?prefix f

	  
let file_iterator_of_channel ?(nodescent=false) ?prefix ch =
  let f() =
    try
      Some(ch # input_line())
    with
      | End_of_file ->
	  None
  in
  if nodescent then
    f
  else
    expand_directories ?prefix f


let next_file f =
  f()


type copier =
    { cp_postprocess : (string, string -> unit) Hashtbl.t }

let create_copier() =
  { cp_postprocess = Hashtbl.create 50 }

let copy_file ?src_prefix ?(prefer_link=false)
              ?(subst=fun s -> s) ?(preserve_mtime = false)
              ?(overwrite = `Always)
              cp src dir =

  if not (Godi_file.file_is_dir dir) then
    raise(Sys_error ("Not a directory: " ^ dir));

  let ar = new Pax_clipboard.clipboard src_prefix in
  let out_ar =
    { out_ar = ar;
      out_wr = ar # create_archive();
      out_close = fun() ->()
    } in
  add_element_from_file ?prefix:src_prefix ~subst out_ar src;
  let in_el_opt = ar # read_archive() in
  match in_el_opt with
    | None -> assert false
    | Some in_el ->
	let in_ar =
	  { in_ar = ar;
	    in_next = (fun () -> None);
	    in_postprocess = cp.cp_postprocess;
	    in_close = fun () -> ()
	  } in
	extract_element_as_file
	  ~insecure:true ~prefer_link
	  ~prefix:dir ~preserve_mtime ~overwrite (in_ar,in_el)
	  

let postprocess_copying cp =
  postprocess_ht cp.cp_postprocess
